; =============================
;  Fellow Amiga Emulator      
;  680X0 emulation core            
;  (C) 1997/1998 Petter Schau 
;            and Roman Dolejsi
; =============================

	        .586p
DGROUP group _DATA,_BSS
_TEXT  segment para public use32 'CODE'
       assume cs:_TEXT, ds:DGROUP, ss:DGROUP

; Reference point for code alignment

startlbl:

include defs.inc
include func\memory.inc
include func\68000.inc

include mac\memory.mac

#ifdef CPU_LOG_INSTRUCTION_START
extern cpuLogInstructionStart_:near
#endif

include mac\68000.mac


;
;                     68030 specific address mode routines
;

;
; Adressmode routines for reading byte,word,longword
; With PC increase
; Parameters:  In:  edx - pointer to register   esi - current PC position
;             Out:  edx - result                esi - current PC position
;

; disp(Ax,Ri) read (68030 compliant)
; ----------------------------------


		ALIGN32

public arl063_
arl063_:	call	eac063_
		mov	ecx, edx
		fetchlret


		ALIGN32

public arw063_
arw063_:	call	eac063_
		mov	ecx, edx
		fetchwret


		ALIGN32

public arb063_
arb063_:	call	eac063_
		mov	ecx, edx
		fetchbret


		ALIGN32

public arl0733_
arl0733_:	call	eac0733_
		mov	ecx, edx
		fetchlret


		ALIGN32

public arw0733_
arw0733_:	call	eac0733_
		mov	ecx, edx
		fetchwret


		ALIGN32

public arb0733_
arb0733_:	call	eac0733_
		mov	ecx, edx
		fetchbret


;
; Adressmode routines for reading byte,word,longword
; (NOTE!) _Without_ PC increase, without An inc/dec
; Parameters:  In:  edx - pointer to register   esi - current PC position
;             Out:  edx - result                esi - current PC position
;

; disp(Ax,Ri) read (68030 compliant)
; ----------------------------------

		ALIGN32

public parl063_
parl063_:	push	esi
		call	eac063_
		pop	esi
		mov	ecx, edx
		fetchlret


		ALIGN32

public parw063_
parw063_:	push	esi
		call	eac063_
		pop	esi
		mov	ecx, edx
		fetchwret


		ALIGN32

public parb063_
parb063_:	push	esi
		call	eac063_
		pop	esi
		mov	ecx, edx
		fetchbret


		ALIGN32

public parl0733_
parl0733_:	push	esi
		call	eac0733_
		pop	esi
		mov	ecx, edx
		fetchlret


		ALIGN32

public parw0733_
parw0733_:	push	esi
		call	eac0733_
		pop	esi
		mov	ecx, edx
		fetchwret


		ALIGN32

public parb0733_
parb0733_:	push	esi
		call	eac0733_
		pop	esi
		mov	ecx, edx
		fetchbret


;
; Adressmode routines for writing byte,word,longword
; With PC increase
; Parameters:  In:  ebx - pointer to register   esi - current PC position
;                   edx - data
;

; disp(Ax,Ri) write (68030 compliant)
; -----------------------------------



		ALIGN32

public awl063_
awl063_:	push	edx
		mov	edx, ebx
		call	eac063_
		mov	ecx, edx
		pop	edx
		writelret


		ALIGN32

public aww063_
aww063_:	push	edx
		mov	edx, ebx
		call	eac063_
		mov	ecx, edx
		pop	edx
		writewret


		ALIGN32

public awb063_
awb063_:	push	edx
		mov	edx, ebx
		call	eac063_
		mov	ecx, edx
		pop	edx
		writebret


;
; Adressmode routines for getting effective address
; With PC increase
; Parameters:  In:  edx - pointer to register   esi - current PC position
;             Out:  edx - result                esi - current PC position
;

; disp(Ax,Ri) get <ea> (68030 compliant)
; --------------------------------------

		ALIGN32

public eac063_
eac063_:	push	ebx
		mov	ebx, dword ptr [edx]	; Get An value
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		test	dh, 1			; extended params (68030)
		jnz	eac063ext
		call	eac063normal
		mov	edx, ecx
		pop	ebx
		ret

eac063ext:	call	eac063extended
		mov	edx, ecx
		pop	ebx
		ret


; disp(pc,Ri) get <ea> (68030 compliant)
; --------------------------------------

		ALIGN32

public eac0733_
eac0733_:	push	ebx
		mov	ebx, esi	; Get PC value
		PTR_TO_PC	ebx
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		test	dh, 1			; extended params (68030)
		jnz	eac0733ext
		call	eac063normal
		mov	edx, ecx
		pop	ebx
		ret

eac0733ext:	call	eac063extended
		mov	edx, ecx
		pop	ebx
		ret


;======================================================
; Brief format (68030) ea calculation
;
; In:	esi - current PC (points after extension word)
;       ebx - Ax or PC value
;	edx - Extension word
; Out:	ecx - Effective Address
;=============================================


		ALIGN32

eac063normal:	movsx	ecx, dl
		add	ecx, ebx
		mov	ebx, edx
		shr	ebx, 10
		and	ebx, 03ch
		mov	ebx, dword ptr [_d + ebx]
		test	dh, 8
		jnz	eac063nosext
		movsx	ebx, bx
eac063nosext:	xchg	edx, ecx
		shr	ecx, 9
		and	ecx, 3
		shl	ebx, cl			; Ri index multiply (68030)
		xchg	edx, ecx
		add	ecx, ebx
		ret


;======================================================
; Extended format (68030) ea calculation
;
; In:	esi - current PC (points after extension word)
;       ebx - Ax or PC value (base register value)
;	edx - Extension word
; Out:	ecx - Effective Address
;======================================================


		ALIGN32

eac063extended: CPU_LOG 063
                push    eax
		test	dl, 80h		; Base register included?
		jz	eac063a
		xor	ebx, ebx	; Base register not included!
eac063a:	mov	eax, edx	; Find out about base displacement
		and	eax, 030h
		jz	eac063illegal	; 0 - Illegal
		cmp	eax, 010h	; 1 - Null base displacement
		je	eac063c
		cmp	eax, 020h	; 2 - Word base displacement
		jne	eac063b
		mov	ecx, esi
		push	edx
		fetchwopcode_inc_pos
		add	esi, 2
		movsx	edx, dx
		add	ebx, edx
		pop	edx
		jmp	eac063c
eac063b:	mov	ecx, esi	; 3 - Long base displacement
		push	edx
		fetchlopcode_inc_pos
		add	esi, 4
		add	ebx, edx
		pop	edx
eac063c:	xor	eax, eax	; ebx = (An) + bd
		test	edx, 040h	; Find out about index
		jnz	eac063e
		mov	al, dh		; Index is included
		shr	al, 2
		and	eax, 3ch
		mov	eax, dword ptr [_d + eax] ; Get index reg value
		test	dh, 8		; Index size check
		jnz	eac063d
		movsx	eax, ax
eac063d:	mov	cl, dh
		shr	cl, 1
		and	cl, 3
		shl	eax, cl		; Multiply index value
eac063e:	mov	ebp, edx	; eax - (Ri)*scale, ebx - (An) + bd
		and	ebp, 07h	; Find IS mode
		cmp	ebp, 3
		jbe	eac063f
		test	edx, 040h	; Test for reserved IS modes
		jnz	eac063illegal
eac063f:	call	dword ptr [eac063istab + 4*ebp] ; Decode IS field
		pop	eax
		ret

eac063illegal:	mov	ebx, 010h	; Reserved, ie. INVALID OPCODE
		mov	eax, 38
		jmp	prepare_exception_


;========================
; IS decode routines
; In:	eax - (Ri)*scale
;       ebx - (An) + bd
;       esi - Current PC
; Out:	ecx - EA
;========================

		ALIGN32
	
eac063is0:	mov	ecx, ebx	; No memory indirect action
		add	ecx, eax	; ie. (An, Ri*scale, bd) 
		ret


		ALIGN32

eac063is1:	mov	ecx, ebx	; Preindexed with null outer disp 
		add	ecx, eax	; ie. ([An, Ri*scale, bd])
		fetchl
		mov	ecx, edx
		ret


		ALIGN32

eac063is2:	mov	ecx, ebx	; Preindexed with word outer disp
		add	ecx, eax	; ie. ([An, Ri*scale, bd], od)
		fetchl			; Fetch intermediary address
		mov	ebx, edx
		mov	ecx, esi	; Fetch outer displacement
		fetchwopcode_inc_pos
		add	esi, 2
		movsx	ecx, dx
		add	ecx, ebx
		ret


		ALIGN32

eac063is3:	mov	ecx, ebx	; Preindexed with long outer disp
		add	ecx, eax	; ie. ([An, Ri*scale, bd], od)
		fetchl			; Fetch intermediary address
		mov	ebx, edx
		mov	ecx, esi	; Fetch outer displacement
		fetchlopcode_inc_pos
		add	esi, 4
		mov	ecx, edx
		add	ecx, ebx
		ret


		ALIGN32

eac063is4:	jmp	eac063illegal	; Reserved, ie. ILLEGAL OPCODE


		ALIGN32

eac063is5:	mov	ecx, ebx	; Postindexed with null outer disp
		fetchl			; Fetch intermediary address
		mov	ecx, edx	; ie. ([An, bd], Ri*scale) 
		add	ecx, eax
		ret


		ALIGN32

eac063is6:	mov	ecx, ebx	; Postindexed with word outer disp
		fetchl			; Fetch intermediary address
		mov	ebx, edx	; ie. ([An, bd], Ri*scale, od)
		add	ebx, eax
		mov	ecx, esi	; Fetch outer displacement
		fetchwopcode_inc_pos
		add	esi, 2
		movsx	ecx, dx
		add	ecx, ebx
		ret


		ALIGN32

eac063is7:	mov	ecx, ebx	; Postindexed with long outer disp
		fetchl			; Fetch intermediary address
		mov	ebx, edx	; ie. ([An, bd], Ri*scale, od)
		add	ebx, eax
		mov	ecx, esi	; Fetch outer displacement
		fetchlopcode_inc_pos
		add	esi, 4
		mov	ecx, edx
		add	ecx, ebx
		ret


;
;                        68030 specific instructions
;

; Common code for bitfields: (Saves 700 lines of sourcecode.)

BF_GET_EXTENSION macro EXTREG
		mov	ecx,esi
		fetchwopcode_inc_pos
		add	esi,2
		mov	EXTREG, edx
endm

BF_GET_SIZE macro EXTREG, SIZEREG
local l1
		mov	SIZEREG, EXTREG
		test	SIZEREG, 020h
		jz	l1
		and	SIZEREG, 7	; size is in Dn 
		mov	SIZEREG, dword ptr [_d + 4*SIZEREG]
l1:		dec	SIZEREG
		and	SIZEREG, 1fh
endm

BF_GET_MASK macro SIZEREG, MASKREG
		mov	ecx, SIZEREG
		mov	MASKREG, 07fffffffh
		shr	MASKREG, cl
		not	MASKREG
endm

BF_GET_SIZEMASK macro EXTREG, MASKREG
		BF_GET_SIZE EXTREG, MASKREG
		BF_GET_MASK MASKREG, MASKREG
endm

BF_GET_OFFSET macro EXTREG, OFFREG
local l1
		mov	OFFREG, EXTREG
		shr	OFFREG, 6
		and	OFFREG, 01fh
		test	EXTREG, 0800h
		jz	l1
		and	OFFREG, 7	; offset is in Dn
		mov	OFFREG, dword ptr [_d + 4*OFFREG]
l1:
endm

BF_FETCH_MEM macro OFFREG, DEST1, DEST2
		mov	DEST2, OFFREG
		sar	DEST2, 3
		and	DEST2, 0fffffffch
		push	DEST2
		mov	edx, dword ptr [4 + edi]
		call	dword ptr [8 + edi]
		pop	ecx
		add	ecx, edx
		mov	ecx, DEST2
		call	fetl
		mov	DEST1, edx
		add	ecx, 4
		call	fetl
		mov	DEST2, edx
endm

BF_FETCH_MEM_EA_SAVE macro OFFREG, DEST1, DEST2
		BF_FETCH_MEM OFFREG, DEST1, DEST2
		push	ecx
endm

BF_FETCH_REG macro OFFREG, DEST2
		mov	DEST2, dword ptr [4 + edi]
		mov	DEST2, dword ptr [DEST2]
		and	OFFREG, 01fh	; ?
endm

BF_FETCH_REG_EA_SAVE macro OFFREG, DEST2
		mov	DEST2, dword ptr [4 + edi]
		push	DEST2
		mov	DEST2, dword ptr [DEST2]
		and	OFFREG, 01fh	; ?
endm

; Offset goes to ecx, and OFFREG is used as the left end of the workspace
BF_SHIFTL_96 macro OFFREG, DEST1, DEST2
		mov	ecx, OFFREG		; length of the shift
		and	ecx, 1fh
		shld	OFFREG, DEST1, cl	; move 96-bit work area
		shld	DEST1, DEST2, cl
		shl	DEST2, cl
endm

BF_SHIFTR_96 macro DEST0, DEST1, DEST2
		shrd	DEST2, DEST1, cl
		shrd	DEST1, DEST0, cl
endm

; Offset goes to ecx, and OFFREG is used as the left end of the workspace
BF_SHIFTL_64 macro OFFREG, DEST1
		mov	ecx, OFFREG		; length of the shift
		and	ecx, 1fh
		shld	OFFREG, DEST1, cl	; move 64-bit work area
		shl	DEST1, cl
endm

BF_SHIFTR_64 macro DEST0, DEST1
		shrd	DEST1, DEST0, cl
endm

BF_DESTREG_PUSH macro EXTREG
		mov	ecx, EXTREG
		shr	ecx, 10
		and	ecx, 01ch
		push	ecx
endm
		
BF_SET_DESTREG_CYCLES macro FIELDREG
		pop	ecx
		mov	dword ptr [_d + ecx], FIELDREG
		mov	eax, dword ptr [12 + edi]
endm

BF_FLAGS macro FREG, MASKREG
local l1, l2
		and	dword ptr [_sr], 0fffffff0h
		test	FREG, 80000000h
		jz	l1
		or	byte ptr [_sr], 8 ; set N according to highest one
l1:		test	FREG, MASKREG
		jnz	l2
		or	byte ptr [_sr], 4 ; set Z according to whole field
l2:
endm

BF_STORE_MEM_CYCLES macro DEST1, DEST2
		pop     ecx
		call    wril
		mov     edx, DEST1
		sub     ecx, 4
		call    wril
		mov     eax, dword ptr [12 + edi]
endm

BF_STORE_REG_CYCLES macro DEST1
		pop     ecx
		mov	dword ptr [ecx], DEST1
		mov     eax, dword ptr [12 + edi]
endm

BF_EXT macro FREG, SHOP
		pop	edx
		mov	cl, 31
		sub	cl, dl
		SHOP	FREG, cl
endm

BF_INS macro FREG, MASKREG
		not	MASKREG
		and	ebp, MASKREG	; Original field cleared 
		not	MASKREG
		push	MASKREG
		mov	MASKREG, dword ptr [8 + esp]
		mov	ecx, 31
		sub	ecx, MASKREG
		mov	MASKREG, dword ptr [12 + esp]
		mov	MASKREG, dword ptr [_d + MASKREG]
		shl	MASKREG, cl
		or	FREG, MASKREG
		pop	MASKREG
endm

BF_FFO macro FREG, MASKREG, DESTREG
local l1
		and	FREG, MASKREG
		xor	MASKREG, MASKREG
		bsr	DESTREG, FREG
		jz	l1
		mov	MASKREG, 31
		sub	MASKREG, DESTREG
l1:		pop	DESTREG
		add	DESTREG, MASKREG
endm


;
; Instruction BFCHG <ea>{offset:size}
; Table use: 0 - routine, 1 - srcreg, 2 - srcproc, 3 - timing
; XX2346
;


		ALIGN32

public i301_M_
i301_M_:	CPU_LOG	301
		BF_GET_EXTENSION ebp
		BF_GET_SIZEMASK ebp, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_MEM_EA_SAVE ebx, ebp, edx
		BF_SHIFTL_96 ebx, ebp, edx
		BF_FLAGS ebp, eax
		xor	ebp, eax
		BF_SHIFTR_96 ebx, ebp, edx
		BF_STORE_MEM_CYCLES ebp, edx
		ret


		ALIGN32

public i301_R_
i301_R_:	CPU_LOG 301
		BF_GET_EXTENSION ebp
		BF_GET_SIZEMASK ebp, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_REG_EA_SAVE ebx, ebp
		BF_SHIFTL_64 ebx, ebp
		BF_FLAGS ebp, eax
		xor	ebp, eax
		BF_SHIFTR_64 ebx, ebp
		BF_STORE_REG_CYCLES ebp, edx
		ret


;
; Instruction BFCLR <ea>{offset:size}
; Table use: 0 - routine, 1 - srcreg, 2 - srcproc, 3 - timing
; XX2346
;


		ALIGN32

public i302_M_
i302_M_:	CPU_LOG	302
		BF_GET_EXTENSION ebp
		BF_GET_SIZEMASK ebp, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_MEM_EA_SAVE ebx, ebp, edx
		BF_SHIFTL_96 ebx, ebp, edx
		BF_FLAGS ebp, eax
		not	eax
		and	ebp, eax
		BF_SHIFTR_96 ebx, ebp, edx
		BF_STORE_MEM_CYCLES ebp, edx
		ret


		ALIGN32

public i302_R_
i302_R_:	CPU_LOG	302
		BF_GET_EXTENSION ebp
		BF_GET_SIZEMASK ebp, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_REG_EA_SAVE ebx, ebp
		BF_SHIFTL_64 ebx, ebp
		BF_FLAGS ebp, eax
		not	eax
		and	ebp, eax
		BF_SHIFTR_64 ebx, ebp
		BF_STORE_REG_CYCLES ebp, edx
		ret


;
; Instruction BFEXTS <ea>{offset:size},Dn
; Table use: 0 - routine, 1 - srcreg, 2 - srcproc, 3 - timing
; XX2346
;


		ALIGN32

public i303_M_
i303_M_:	CPU_LOG	303
		BF_GET_EXTENSION ebp
		BF_DESTREG_PUSH ebp
		BF_GET_SIZE ebp, eax
		push	eax
		BF_GET_MASK eax, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_MEM ebx, ebp, edx
		BF_SHIFTL_96 ebx, ebp, edx
		BF_FLAGS ebp, eax
		BF_EXT ebp, sar
		BF_SET_DESTREG_CYCLES ebp
		ret
	

		ALIGN32

public i303_R_
i303_R_:	CPU_LOG	303
		BF_GET_EXTENSION ebp
		BF_DESTREG_PUSH ebp
		BF_GET_SIZE ebp, eax
		push	eax
		BF_GET_MASK eax, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_REG ebx, ebp
		BF_SHIFTL_64 ebx, ebp
		BF_FLAGS ebp, eax
		BF_EXT ebp, sar
		BF_SET_DESTREG_CYCLES ebp
		ret


;
; Instruction BFEXTU <ea>{offset:size},Dn
; Table use: 0 - routine, 1 - srcreg, 2 - srcproc, 3 - timing
; XX2346
;


		ALIGN32

public i304_M_
i304_M_:	CPU_LOG	304
		BF_GET_EXTENSION ebp
		BF_DESTREG_PUSH ebp
		BF_GET_SIZE ebp, eax
		push	eax
		BF_GET_MASK eax, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_MEM ebx, ebp, edx
		BF_SHIFTL_96 ebx, ebp, edx
		BF_FLAGS ebp, eax
		BF_EXT ebp, shr
		BF_SET_DESTREG_CYCLES ebp
		ret
	

		ALIGN32

public i304_R_
i304_R_:	CPU_LOG 304
		BF_GET_EXTENSION ebp
		BF_DESTREG_PUSH ebp
		BF_GET_SIZE ebp, eax
		push	eax
		BF_GET_MASK eax, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_REG ebx, ebp
		BF_SHIFTL_64 ebx, ebp
		BF_FLAGS ebp, eax
		BF_EXT ebp, shr
		BF_SET_DESTREG_CYCLES ebp
		ret


;
; Instruction BFFFO <ea>{offset:size},Dn
; Table use: 0 - routine, 1 - srcreg, 2 - srcproc, 3 - timing
; XX2346
;


		ALIGN32

public i305_M_
i305_M_:	CPU_LOG 305
		BF_GET_EXTENSION ebp
		BF_DESTREG_PUSH ebp
		BF_GET_SIZEMASK ebp, eax
		BF_GET_OFFSET ebp, ebx
		push	ebx	; save offset
		BF_FETCH_MEM ebx, ebp, edx
		BF_SHIFTL_96 ebx, ebp, edx
		BF_FLAGS ebp, eax
		BF_FFO ebp, eax, ebx
		BF_SET_DESTREG_CYCLES ebx
		ret


		ALIGN32

public i305_R_
i305_R_:	CPU_LOG 305
		BF_GET_EXTENSION ebp
		BF_DESTREG_PUSH ebp
		BF_GET_SIZEMASK ebp, eax
		BF_GET_OFFSET ebp, ebx
		push	ebx	; save offset
		BF_FETCH_REG ebx, ebp
		BF_SHIFTL_64 ebx, ebp
		BF_FLAGS ebp, eax
		BF_FFO ebp, eax, ebx
		BF_SET_DESTREG_CYCLES ebx
		ret


;
; Instruction BFINS Dn,<ea>{offset:size}
; Table use: 0 - routine, 1 - dstreg, 2 - dstproc, 3 - timing
; XX2346
;


		ALIGN32

public i306_M_
i306_M_:	CPU_LOG	306
		BF_GET_EXTENSION ebp
		BF_DESTREG_PUSH ebp
		BF_GET_SIZE ebp, eax
		push	eax
		BF_GET_MASK eax, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_MEM_EA_SAVE ebx, ebp, edx
		BF_SHIFTL_96 ebx, ebp, edx
		BF_INS ebp, eax
		BF_FLAGS ebp, eax
		BF_SHIFTR_96 ebx, ebp, edx
		BF_STORE_MEM_CYCLES ebp, edx
		pop	ebx	; Garbage on stack
		pop	ecx
		ret
	

		ALIGN32

public i306_R_
i306_R_:	CPU_LOG	306
		BF_GET_EXTENSION ebp
		BF_DESTREG_PUSH ebp
		BF_GET_SIZE ebp, eax
		push	eax
		BF_GET_MASK eax, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_REG_EA_SAVE ebx, ebp
		BF_SHIFTL_64 ebx, ebp
		BF_INS ebp, eax
		BF_FLAGS ebp, eax
		BF_SHIFTR_64 ebx, ebp
		BF_STORE_REG_CYCLES ebp, edx
		pop	ebx	; Garbage on stack
		pop	ecx
		ret


;
; Instruction BFSET <ea>{offset:size}
; Table use: 0 - routine, 1 - srcreg, 2 - srcproc, 3 - timing
; XX2346
;


		ALIGN32

public i307_M_
i307_M_:	CPU_LOG 307
		BF_GET_EXTENSION ebp
		BF_GET_SIZEMASK ebp, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_MEM_EA_SAVE ebx, ebp, edx
		BF_SHIFTL_96 ebx, ebp, edx
		BF_FLAGS ebp, eax
		or	ebp, eax
		BF_SHIFTR_96 ebx, ebp, edx
		BF_STORE_MEM_CYCLES ebp, edx
		ret


		ALIGN32

public i307_R_
i307_R_:	CPU_LOG 307
		BF_GET_EXTENSION ebp
		BF_GET_SIZEMASK ebp, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_REG_EA_SAVE ebx, ebp
		BF_SHIFTL_64 ebx, ebp
		BF_FLAGS ebp, eax
		or	ebp, eax
		BF_SHIFTR_64 ebx, ebp
		BF_STORE_REG_CYCLES ebp, edx
		ret

	
;
; Instruction BFTST <ea>{offset:size}
; Table use: 0 - routine, 1 - srcreg, 2 - srcproc, 3 - timing
; XX2346
;


		ALIGN32

public i308_M_
i308_M_:	CPU_LOG	308
		BF_GET_EXTENSION ebp
		BF_GET_SIZEMASK ebp, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_MEM ebx, ebp, edx
		BF_SHIFTL_96 ebx, ebp, edx
		BF_FLAGS ebp, eax
		mov	eax, dword ptr [12 + edi]
		ret


		ALIGN32

public i308_R_
i308_R_:	CPU_LOG	308
		BF_GET_EXTENSION ebp
		BF_GET_SIZEMASK ebp, eax
		BF_GET_OFFSET ebp, ebx
		BF_FETCH_REG ebx, ebp
		BF_SHIFTL_64 ebx, ebp
		BF_FLAGS ebp, eax
		mov	eax, dword ptr [12 + edi]
		ret

	
;
; Instruction CAS Dc,Du,<ea>                                 (All, not X)
; Table use: 0 - routine, 1 - srcreg, 2 - eacproc, 3 - timing
; XX234X
;

CAS macro FETCHFUNC, WRITEFUNC, EBXSIZED, EDXSIZED, SIZESPEC
local l1
		mov	edx, [edi + 4]
		call	[edi + 8]		; edx = <ea>
		mov	eax, edx		; eax is <ea> address 
		mov	ecx, esi
		fetchwopcode_inc_pos		; get extra word
		add	esi, 2
		push	edx			; save extension word
		mov	bl, dl			; Start to get Dc
		shl	bl, 2
		and	ebx, 1ch
		mov	ebx, [_d + ebx]		; ebx holds Dc operand
		mov	ecx, eax		; Start to get value at <ea>
		call	FETCHFUNC
		cmp	EBXSIZED, EDXSIZED	; Compare Dc with <ea> operand
		pushfd				; Set flags
		pop	ecx
		and	ecx, 0fffh
		mov	ebx, dword ptr [_statustab + 4*ecx]
		mov	ecx, dword ptr [_sr]
		and	ecx, 0fffffff0h
		and	ebx, 0fh
		or	ecx, ebx
		mov	dword ptr [_sr], ecx

		cmp	EBXSIZED, EDXSIZED
		jnz	l1

		pop	ebx			; write Du to <ea>
		shr	bx, 4
		and	ebx, 1ch
		mov	edx, [_d + ebx]
		mov	ecx, eax
		call	WRITEFUNC
		mov	eax, [edi + 12]
		ret

l1:		pop	ebx			; write <ea> value to Dc
		shl	bl, 2
		and	ebx, 1ch
		mov	SIZESPEC ptr [_d + ebx], EDXSIZED
		mov	eax, [edi + 12]
		ret
endm


		ALIGN32

public i309_B_
i309_B_:	CPU_LOG	309
		CAS	fetb, wrib, bl, dl, byte


		ALIGN32

public i309_W_
i309_W_:	CPU_LOG 309
		CAS	fetw, wriw, bx, dx, word


		ALIGN32

public i309_L_
i309_L_:	CPU_LOG 309
		CAS	fetl, wril, ebx, edx, dword


;
; Instruction CAS2 Dc1:Dc2,Du1:Du2,(Rx1):(Rx2)              (All, not X)
; Table use: 0 - routine, 1 - timing
; XX234X
;

CAS2 macro FETCHFUNC, WRITEFUNC, EDXSIZED, SIZESPEC
local l1, l2
		mov	ecx, esi
		fetchlopcode_inc_pos		; get params
		add	esi, 4
		mov	eax, edx
		ror	edx, 16			; first we need Dc1/Du1/Rx1
		mov	bx, dx
		rol	bx, 6
		and	ebx, 3ch
		mov	ecx, [_d + ebx]
		call	FETCHFUNC
		push	edx			; Save data from (Rx1)
		push	ecx			; Save eac from Rx1
		mov	ebx, eax
		shr	ebx, 14
		and	ebx, 1ch		; Dc1 register number*4
		cmp	EDXSIZED, SIZESPEC ptr [_d + ebx] ; Dc1 = Rx1 ?
		jne	l1
		mov	bx, ax			; First compare equal
		rol	bx, 6
		and	ebx, 3ch
		mov	ecx, [_d + ebx]		; ecx - rx2 ea
		call	FETCHFUNC		; edx - rx2 value
		mov	bx, ax
		shl	bx, 2
		and	ebx, 1ch		; ebx - dc2 regnr*4
		cmp	EDXSIZED, SIZESPEC ptr [_d + ebx] ; Dc2 = Rx2 ?
		push	ecx
		pushfd				; Set flags for Dc2 Rx2 cmp
		pop     ecx
		and	ecx, 0fffh
		mov	ebx, dword ptr [_statustab+4*ecx]
		mov	ecx, dword ptr _sr
		and	ecx, 0fffffff0h
		and	ebx, 0fh
		or	ecx, ebx
		mov	dword ptr _sr, ecx
		pop	ecx
		cmp	EDXSIZED, SIZESPEC ptr [_d + ebx] ; Dc2 = Rx2 ?
		jne	l2
		mov	bx, ax			; Perform Rx1 = Du1, Rx2 = Du2
		shr	bx, 4
		and	bx, 1ch
		mov	edx, [_d + ebx]		; Get Du2 value
		call	WRITEFUNC		; (Rx2) <- Du2
		pop	ecx			; Reload Rx1 ea
		ror	eax, 16
		mov	bx, ax
		shr	bx, 4
		and	bx, 1ch
		mov	edx, [_d + ebx]		; Get Du1 value
		call	WRITEFUNC		; (Rx1) <- Du1
		pop	edx
		mov	eax, [edi + 4]
		ret

l1:		pushfd				; First cmp failed, set flags
		pop	ecx			; Then do Dc1 = Rx1, Dc2 = Rx2
		and	ecx, 0fffh
		mov	ebx, dword ptr [_statustab+4*ecx]
		mov	ecx, dword ptr _sr
		and	ecx, 0fffffff0h
		and	ebx, 0fh
		or	ecx, ebx
		mov	dword ptr _sr, ecx
l2:		pop	ecx			; eac Rx1, dont need it
		pop	edx			; (Rx1) value
		mov	ebx, eax
		ror	ebx, 20
		and	ebx, 1ch
		mov	SIZESPEC ptr [_d + ebx], EDXSIZED ; Dc1 <- (Rx1)
		mov	bx, ax
		rol	bx, 6
		and	ebx, 3ch
		mov	ecx, [_d + ebx]		; eac Rx2
		call	FETCHFUNC		; may not have Rx2 val
		mov	bx, ax
		rol	bx, 2
		and	ebx, 1ch
		mov	SIZESPEC ptr [_d + ebx], EDXSIZED ; Dc2 <- (Rx2)
		mov	eax, [edi + 4]
		ret
endm


		ALIGN32

public i310_W_
i310_W_:	CPU_LOG	310
		CAS2	fetw, wriw, dx, word


		ALIGN32

public i310_L_
i310_L_:	CPU_LOG 310
		CAS2	fetl, wril, edx, dword


;
; Instruction CHK.L <ea>,Dn
; Table use: 0 - routine, 1 - srcreg
;            2 - srcproc, 3 - dstreg
; NOTE:	Word version in 68000.asm
; Word:	012346
; Long:	XX2346
;


		ALIGN32

public i311_
i311_:		CPU_LOG	311
		mov	eax, dword ptr [edi + 8]
		mov	edx, dword ptr [edi + 4]
		call	eax
		mov	ebx, dword ptr [edi + 12]
		mov	eax, dword ptr [ebx]
		and	dword ptr [_sr], 0fffffffbh
		or	eax, eax
		js	i311trapset
		cmp	eax, edx
		ja	i311trap
		mov	eax, 4
		ret
i311TrapSet:	or	dword ptr [_sr], 8
i311Trap:	mov	ebx, 18h		; vector at $18 (nr. 6)
		mov	eax, 34
		add	eax, 4
		jmp	prepare_exception_


;
; Instruction CHK2.X <ea>,Rn / CMP2.X <ea>,Rn
; Table use: 0 - routine, 1 - srcreg, 2 - srcproc, 3 - timing
; XX234X
;

; Sign extend register is usually EDXSIZED, but since sign extend of
; long to long is not allowed, this is just a dummy for the long version

CHKCMP2 macro EAXSIZED, EDXSIZED, SEXTREG, BINC
local ZFSet, CFSet, End1, End2, Areg
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		mov	ebx, edx
		rol	bx, 6
		push	ebx
		and	ebx, 3ch
		mov	eax, [_d + ebx]			; eax - Rn value 
		push	eax				; Save Rn value
		mov	eax, dword ptr [edi + 8]	; Fetch lower bound 
		mov	edx, dword ptr [edi + 4]
		call	eax
		pop	eax
		and	dword ptr _sr, 0fffffffah
		test	bl, 20h
		jnz	Areg
		cmp	EAXSIZED, EDXSIZED		; Cmp lower bound
		jz	ZFset
		js	CFset
		push	eax				; Lower OK
		mov	eax, dword ptr [edi + 8]
		mov	edx, dword ptr [edi + 4]
		add	edx, BINC
		call	eax				; Fetch higher bound
		pop	eax
		cmp	EAXSIZED, EDXSIZED		; Cmp higher bound
		ja	CFSet
		jnz	End1
ZFSet:		or	dword ptr _sr, 4
End1:		pop	ebx
End2:		mov	eax, dword ptr [edi + 12]
		ret
CFSet:		or	dword ptr _sr, 1
		pop	ebx
		test	bl, 2			; are we doing CHK2 or CMP2 ?
		jz	End2
		mov	ebx, 18h		; vector at $18 (nr. 6)
		mov	eax, 34
		add	eax, dword ptr [edi + 12]
		jmp	prepare_exception_

Areg:		movsx	edx, SEXTREG		; Sign extend
		cmp	eax, edx
		jz	ZFset
		js	CFset
		push	eax
		mov	eax, dword ptr [edi + 8]
		mov	edx, dword ptr [edi + 4]
		add	edx, BINC
		call	eax
		pop	eax
		movsx	edx, SEXTREG
		cmp	eax, edx
		ja	CFSet
		jnz	End1
		or	dword ptr _sr, 4
		pop	ebx
		mov	eax, dword ptr [edi + 12]
		ret
endm


		ALIGN32

public i312_B_
i312_B_:	CPU_LOG	312
		CHKCMP2 al, dl, dl, 2


		ALIGN32

public i312_W_
i312_W_:	CPU_LOG	312
		CHKCMP2	ax, dx, dl, 2


		ALIGN32

public i312_L_
i312_L_:	CPU_LOG 312
		CHKCMP2	eax, edx, bl, 4


;
; Instruction cpBcc <label>
; Table use: 0 - routine, 1 - disasm, 2 - coprocessor, 3 - condition
;

extern i00_:near
	
cpillegal:	jmp	i00_



		ALIGN32

public i313_W_
i313_W_:	CPU_LOG	313
		jmp	cpillegal
	
	
	
	push    dword ptr [edi + 12]            ; prepare condition
        push    offset i313_si                  ; store PC reader
        push    edi                             ; prepare table ptr
        push    esi                             ; prepare current PC
        mov     ebx, [edi + 8]
        shl     ebx, 4                          ; table with 4 procs per copr
        call    [_coproc_tbl + ebx]             ; coprproc0
        pop     esi
        pop     edi
        pop     ecx
        pop     ecx
        movzx   ecx, ax                         ; PC skip in low word
        add     esi, ecx                        ; inc PC (skip copr instrs)
        test    eax, 10000h
        jz      i313_Wn
        mov     ecx, esi
        fetchw
        movsx   edx, dx
        add     esi, edx
i313_Wn:add     esi, 2
        mov     eax, 0
        ret


		ALIGN32

public i313_L_
i313_L_:	CPU_LOG	313
		jmp	cpillegal

	push    dword ptr [edi + 12]            ; prepare condition
        push    offset i313_si                  ; store PC reader
        push    edi                             ; prepare table ptr
        push    esi                             ; prepare current PC
        mov     ebx, [edi + 8]
        shl     ebx, 4                          ; table with 4 procs per copr
        call    [_coproc_tbl + ebx]             ; coprproc0
        pop     esi
        pop     edi
        pop     ecx
        pop     ecx
        movzx   ecx, ax                         ; PC skip in low word
        add     esi, ecx                        ; inc PC (skip copr instrs)
        test    eax, 10000h
        jz      i313_Ln                         ; condition not met
        mov     ecx, esi
        fetchl
        add     esi, edx
i313_Ln:add     esi, 4
        mov     eax, 0
        ret

i313_si:push    ebp
        mov     ebp, esp
        mov     ecx, [ebp + 4]
        fetchw
        pop     ebp
        ret


;
; Instruction cpDBcc <label>
; Table use: 0 - routine, 1 - disasm, 2 - coprocessor, 3 - *Dn
;


		ALIGN32

public i314_
i314_:		CPU_LOG	314
		jmp	cpillegal

	  mov     ecx, esi
        fetchw
        add     esi, 2
        and     edx, 3fh
        push    edx                             ; prepare condition
        push    offset i313_si                  ; store PC reader
        push    edi                             ; prepare table ptr
        push    esi                             ; prepare current PC
        mov     ebx, [edi + 8]
        shl     ebx, 4                          ; table with 4 procs per copr
        call    [_coproc_tbl + ebx]             ; coprproc0
        pop     esi
        pop     edi
        pop     edx
        pop     edx

        movzx   ecx, ax                         ; PC skip in low word
        add     esi, ecx                        ; inc PC (skip copr instrs)
        test    eax, 10000h
        jnz     i313_Ln                         ; inverted result!
        mov     ebx, [edi + 12]
        dec     word ptr [ebx]                  ; dec Dn
        jc      i314_n                          ; end of loop ?
        mov     ecx, esi
        fetchw
        movsx   edx, dx
        add     esi, edx
i314_n: add     esi, 2
        mov     eax, 0
        ret


;
; Instruction cpGEN
; Table use: 0 - routine, 1 - disasm, 2 - coprocessor, 3 - reg, 4 - srcproc
;


		ALIGN32

public i315_
i315_:		CPU_LOG	315
		jmp	cpillegal

	  mov     ecx, esi
        fetchw
        add     esi, 2
	ret

	
        push    edx                             ; prepare command word
        push    offset i313_si                  ; store PC reader
        push    offset i315_rd                  ; store ea reading proc
        push    offset i317_wr                  ; store ea writing proc
        push    edi                             ; prepare table ptr
        push    esi                             ; prepare current PC
        mov     ebx, [edi + 8]
        shl     ebx, 4                          ; table with 4 procs per copr
        call    [_coproc_tbl + 4 + ebx]         ; coprproc1
        pop     esi
        pop     edi
        pop     edx
        pop     edx
        pop     edx
        pop     edx
        movzx   eax, ax
        add     esi, eax                        ; inc PC (skip copr instrs)
        mov     eax, 0
        ret

i315_rd:push    ebp
        mov     ebp, esp
        mov     edi, [ebp + 4]
        mov     edx, [edi + 12]
        call    [edi + 16]
        mov     eax, edx
        pop     ebp
        ret


;
; Instruction cpRESTORE
; Table use: 0 - routine, 1 - disasm, 2 - coprocessor, 3 - reg, 4 -srcproc
;


		ALIGN32

public i316_
i316_:		CPU_LOG	316
	; Note supervisor
		jmp	cpillegal

	  push    offset i315_rd                  ; store reading proc
        push    edi                             ; prepare table ptr
        push    esi                             ; prepare current PC
        mov     ebx, [edi + 8]
        shl     ebx, 4                          ; table with 4 procs per copr
        call    [_coproc_tbl + 8 + ebx]         ; coprproc2
        pop     esi
        pop     edi
        pop     edx
        movzx   eax, ax
        add     esi, eax                        ; inc PC (skip copr instrs)
        mov     eax, 0
        ret


;
; Instruction cpSAVE
; Table use: 0 - routine, 1 - disasm, 2 - coprocessor, 3 - reg, 4 -dstreg
;


		ALIGN32

public i317_
i317_:		CPU_LOG	317
					; NOTE supervisor
		jmp	cpillegal

	  push    offset i317_wr                  ; store writing proc
        push    edi                             ; prepare table ptr
        push    esi                             ; prepare current PC
        mov     ebx, [edi + 8]
        shl     ebx, 4                          ; table with 4 procs per copr
        call    [_coproc_tbl + 12 + ebx]        ; coprproc3
        pop     esi
        pop     edi
        pop     edx
        movzx   eax, ax
        add     esi, eax                        ; inc PC (skip copr instrs)
        mov     eax, 0
        ret

i317_wr:push    ebp
        mov     ebp, esp
        mov     edi, [ebp + 8]
        mov     edx, [ebp + 4]
        mov     ebx, [edi + 12]
        call    [edi + 16]
        pop     ebp
        ret


;
; Instruction cpScc
; Table use: 0 - routine, 1 - disasm, 2 - coprocessor, 3 - reg, 4 - dstproc
;


		ALIGN32

public i318_
i318_:		CPU_LOG	318
		jmp	cpillegal

	  mov     ecx, esi
        fetchw
        add     esi, 2
        and     edx, 3fh
        push    edx                             ; prepare condition
        push    offset i313_si                  ; store PC reader
        push    edi                             ; prepare table ptr
        push    esi                             ; prepare current PC
        mov     ebx, [edi + 8]
        shl     ebx, 4                          ; table with 4 procs per copr
        call    [_coproc_tbl + ebx]             ; coprproc0
        pop     esi
        pop     edi
        pop     ecx
        pop     ecx
        movzx   ecx, ax
        add     esi, eax                        ; inc PC (skip copr instrs)
        mov     dl, 0
        test    eax, 10000h
        jz      i318_1
        dec     dl
i318_1: mov     ebx, [edi + 12]
        call    [edi + 16]
        mov     eax, 4
        ret


;
; Instruction cpTRAPcc
; Table use: 0 - routine, 1 - disasm, 2 - coprocessor, 3 - optype (00/1W/2L)
;


		ALIGN32

public i319_
i319_:		CPU_LOG	319
		jmp	cpillegal

	  mov     ecx, esi
        fetchw
        add     esi, 2
        and     edx, 3fh
        push    edx                             ; prepare condition
        push    offset i313_si                  ; store PC reader
        push    edi                             ; prepare table ptr
        push    esi                             ; prepare current PC
        mov     ebx, [edi + 8]
        shl     ebx, 4                          ; table with 4 procs per copr
        call    [_coproc_tbl + ebx]             ; coprproc0
        pop     esi
        pop     edi
        pop     ecx
        pop     ecx
        movzx   ecx, ax
        add     esi, eax                        ; inc PC (skip copr instrs)
        test    eax, 10000h
        jnz     i319_n                          ; condition met
        mov     eax, 0
        ret

i319_n: cmp     dword ptr [edi + 12], 0
        je      i319_0
        mov     ecx, esi
        cmp     dword ptr [edi + 12], 1
        je      i319_1
        fetchl
        add     esi, 4
        jmp     short i319_e
i319_1: fetchw
        add     esi, 2
i319_e:                                         ; here should come
                                                ; trap variable storing!
i319_0: mov     ebx, 7*4 + 80h
	mov     eax, 4
        jmp     prepare_exception_


;
; Instruction DIVS.L <ea>,Dq / <ea>,Dr:Dq / DIVSL <ea>,Dr:Dq       (NZV,C=0)
; Table use: 0 - routine, 1 - ea register, 2 - ea read routine
; 32-bit: XX2346
; 64-bit: XX234X
;

extern i320_C_:near

; Convert all types of divisions to 64/32 bits.
; For 68k 64 bit divisions, an overflow can occur.
; Problem: Intel generates an exception on divide overflows.
;          To predict overflow in the result, do the following (64bit) check:
;          if (MAXINT32*(divisor+1) <= quotient) then overflow
; In the resulting conversion to signed numbers, an overflow can occur,
; does M68K treat this as a proper overflow and leave the operands untouched?


; 1. Get extension word
; 2. Get Dq, quotient register nr and value, save register nr
; 3. Get Dr, remainder register nr, save register nr
; 4. Read divisor from <ea>
; 5. Trap on zero divisor
; 6.1 If 64-bit then Get high long word from Dr
; 6.2 else if (DIVS) sign exted Dq else zero extend Dq
; 7.1 If (DIVS) determine sign of result, convert to unsigned
; 7.2 else sign is positive
; 8. Unsigned divide 
; 9. If (DIVS) convert result to signed, detect overflow
; 10.1 Store result, if (64-bit) put in DrDq pair.
; 10.2 if (32-bit) store remainder in Dr, then result in Dq.
; 
; 
; 

; Note:	Step 8, expected sign of result stored in bit 15 of ebp

NEG64 macro
local l1
		neg	eax
		jnc	l1
		inc	edx
l1:		neg	edx
endm

; Divisons are simply too messy to bother with it in asm. This one reads the
; extension word, the ea value and leaves the rest to a C routine.
	

		ALIGN32

public i320_
i320_:		CPU_LOG	320
		mov	ecx, esi			; step 1 Ext. word
		fetchwopcode_inc_pos
		add	esi, 2
		mov	eax, edx
		mov	edx, dword ptr [4 + edi]	; step 4 Get ea
		call	dword ptr [8 + edi]
		push	esi
		call	i320_C_
		pop	esi
		test	eax, eax
		jz	i320_div0
		js	i320_unimp_060
		mov	eax, 4
		ret

i320_div0:	mov	ebx, 14h
		mov	eax, 200
		jmp	prepare_exception_


i320_unimp_060:			; Raise exception
		ret
	
;
; Instruction EXTB Dn
; Table use: 0 - Routine pointer, 1 - Register pointer
; Flag: X unaffected, V = C = 0, NZ
; Byte:	XX2346
;


		ALIGN32

public i321_
i321_:		CPU_LOG	321
		mov	ecx, dword ptr [edi + 4]
		mov	eax, 4
		mov	edx, dword ptr [_sr]
		and	edx, 0fffffff0h
		mov	ebx, dword ptr [ecx]
		movsx	ebx, bl
		mov	dword ptr [ecx], ebx
		test	ebx, ebx
		jz	i321wz
		jns	i321wns
		or	edx, 8
		jmp	i321wns
i321wz:		or	edx, 4
i321wns:	mov	dword ptr [_sr], edx
		ret


;
; Instruction LINK.L An,#                                          ()
; Table use: 0 - routine, 1 - register pointer
; Word:	012346 found in 68000.asm
; Long:	XX2346
;


		ALIGN32

public i322_
i322_:		CPU_LOG	322
		mov	ebx, [edi + 4]
		mov	edx, [ebx]                 ; get An
		mov	ecx, [_a + 28]
		sub	ecx, 4
		mov	[ebx], ecx                 ; Store new A7 into An
		writel                             ; (A7) <- An
		mov	ebx, ecx
		mov	ecx, esi
		fetchlopcode_inc_pos
		add	esi, 4
		add	ebx, edx
		mov	[_a+28], ebx               ; A7 += #
		mov	eax, 16
		ret


;
; Instruction MOVE CCR,<ea>                                        ()
; Table use: 0 - routine, 1 - dstreg, 2 - dstproc, 3 - timing
; X12346
;


		ALIGN32

public i323_
i323_:		CPU_LOG	323
		mov	dl, byte ptr [_sr]
		mov	dh, 0
		mov	ebx, [edi + 4]
		call	[edi + 8]
		mov	eax, dword ptr [edi + 12]
		ret


;
; Instruction MOVEC Rc,Rn / Rn,Rc                                  ()
; Table use: 0 - routine
; X12346
;

; Not speed-critical, cleanness favored over speed.

; Move to cacr, collect and report dummy values


		ALIGN32

cacrTo020_:	and	eax, 3
		mov	dword ptr [_cacr], eax
		ret


		ALIGN32

cacrTo030_:	and	eax, 0303h
		mov	dword ptr [_cacr], eax
		ret


		ALIGN32

cacrTo040_:	and	eax, 080008000h
		mov	dword ptr [_cacr], eax
		ret


		ALIGN32

cacrTo060_:	and	eax, 0f880e000h
		mov	dword ptr [_cacr], eax
		ret


		ALIGN32

public i324_TO_
i324_TO_:	CPU_LOG	324
		test	byte ptr [_sr + 1], 20h	; Are we in supervisormode??
		jz	privilegie_violation
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		mov	ebx, edx
		mov	eax, edx
		and	ebx, 0fffh
		shr	eax, 10
		and	eax, 03ch
		mov	eax, dword ptr [_d + eax]

		cmp	ebx, 000h		; SFC  X12346
		jne	i324w_1
		and	eax, 7 
		mov	dword ptr [_sfc], eax
		jmp	i324w_end

i324w_1:	cmp	ebx, 001h		; DFC  X12346
		jne	i324w_2
		and	eax, 7
		mov	dword ptr [_dfc], eax
		jmp	i324w_end

i324w_2:	cmp	ebx, 0800h		; USP  X12346
		jne	i324w_3
		mov	dword ptr [_usp], eax
		jmp	i324w_end

i324w_3:	cmp	ebx, 0801h		; VBR  X12346
		jne	i324w_4
		mov	dword ptr [_vbr], eax
		jmp	i324w_end

i324w_4:	cmp	ebx, 002h		; CACR XX2346
		jne	i324w_5

		cmp	dword ptr [_config_cpu_type], 2
		jb	i324w_illegal
		jne	i324w_11
		call	cacrTo020_
		jmp	i324w_end
i324w_11:	cmp	dword ptr [_config_cpu_type], 3
		jne	i324w_12
		call	cacrTo030_
		jmp	i324w_end
i324w_12:	cmp	dword ptr [_config_cpu_type], 4
		jne	i324w_13
		call	cacrTo040_
		jmp	i324w_end
i324w_13:	call	cacrTo060_
		jmp	i324w_end

i324w_5:	cmp	ebx, 0802h		; CAAR XX23XX
		jne	i324w_6

		cmp	dword ptr [_config_cpu_type], 2
		jb	i324w_illegal
		cmp	dword ptr [_config_cpu_type], 4
		jae	i324w_illegal
		and	eax, 0fch
		mov	dword ptr [_caar], eax
		jmp	i324w_end

i324w_6:	cmp	ebx, 0803h		; MSP  XX234X
		jne	i324w_7

		cmp	dword ptr [_config_cpu_type], 2
		jb	i324w_illegal
		cmp	dword ptr [_config_cpu_type], 6
		jae	i324w_illegal
		mov	dword ptr [_msp], eax
		jmp	i324w_end

i324w_7:	cmp	ebx, 0804h		; ISP(SSP here)  XX2346
		jne	i324w_8

		cmp	dword ptr [_config_cpu_type], 2
		jb	i324w_illegal
		mov	dword ptr [_ssp], eax
		jmp	i324w_end

i324w_8:	cmp	ebx, 004h		; IACR0 (EC040)
		jne	i324w_14
		cmp	dword ptr [_config_cpu_type], 4
		jne	i324w_illegal
	
		jmp	i324w_end
i324w_14:	cmp	ebx, 005h		; IACR1 (EC040)
		jne	i324w_15
		cmp	dword ptr [_config_cpu_type], 4
		jne	i324w_illegal

		jmp	i324w_end
i324w_15:	cmp	ebx, 006h		; DACR0 (EC040)
		jne	i324w_16
		cmp	dword ptr [_config_cpu_type], 4
		jne	i324w_illegal

		jmp	i324w_end
i324w_16:	cmp	ebx, 007h		; DACR1 (EC040)
		jne	i324w_17
		cmp	dword ptr [_config_cpu_type], 4
		jne	i324w_illegal

		jmp	i324w_end
i324w_17:					; ILLEGAL OPCODE when here
i324w_illegal:	mov	ebx,010h
		mov	eax,38
		jmp	prepare_exception_

i324w_end:	mov	eax, 4
		ret

; From control register


		ALIGN32

public i324_FROM_
i324_FROM_:	CPU_LOG	324
		test	byte ptr _sr + 1, 20h	; Are we in supervisormode??
		jz	privilegie_violation
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		mov	ebx, edx
		mov	eax, edx
		and	ebx, 0fffh
		shr	eax, 10
		and	eax, 03ch

		cmp	ebx, 000h		; SFC  X12346
		jne	i324m_1
		mov	ecx, dword ptr [_sfc]
		mov	dword ptr [_d + eax], ecx
		jmp	i324m_end

i324m_1:	cmp	ebx, 001h		; DFC  X12346
		jne	i324m_2
		mov	dword ptr [_dfc], eax
		mov	dword ptr [_d + eax], ecx
		jmp	i324m_end

i324m_2:	cmp	ebx, 0800h		; USP  X12346
		jne	i324m_3
		mov	eax, dword ptr [_usp]
		mov	dword ptr [_d + eax], ecx
		jmp	i324m_end

i324m_3:	cmp	ebx, 0801h		; VBR  X12346
		jne	i324m_4
		mov	ecx, dword ptr [_vbr]
		mov	dword ptr [_d + eax], ecx
		jmp	i324m_end

i324m_4:	cmp	ebx, 002h		; CACR XX2346
		jne	i324m_5

		cmp	dword ptr [_config_cpu_type], 2
		jb	i324m_illegal
		mov	ecx, dword ptr [_cacr]
		mov	dword ptr [_d + eax], ecx
		jmp	i324m_end

i324m_5:	cmp	ebx, 0802h		; CAAR XX23XX
		jne	i324m_6

		cmp	dword ptr [_config_cpu_type], 2
		jb	i324m_illegal
		cmp	dword ptr [_config_cpu_type], 4
		jae	i324m_illegal
		mov	ecx, dword ptr [_caar]
		mov	dword ptr [_d + eax], ecx
		jmp	i324m_end

i324m_6:	cmp	ebx, 0803h		; MSP  XX234X
		jne	i324m_7
		cmp	dword ptr [_config_cpu_type], 2
		jb	i324m_illegal
		cmp	dword ptr [_config_cpu_type], 6
		jae	i324m_illegal
		mov	ecx, dword ptr [_msp]
		mov	dword ptr [_d + eax], ecx
		jmp	i324m_end

i324m_7:	cmp	ebx, 0804h		; ISP  XX2346
		jne	i324m_8
		cmp	dword ptr [_config_cpu_type], 2
		jb	i324m_illegal
		mov	ecx, dword ptr [_ssp]
		mov	dword ptr [_d + eax], ecx
		jmp	i324m_end

i324m_8:	cmp	ebx, 004h		; IACR0 (EC040)
		jne	i324m_14
	
		cmp	dword ptr [_config_cpu_type], 4
		jne	i324m_illegal
		jmp	i324m_end
i324m_14:	cmp	ebx, 005h		; IACR1 (EC040)
		jne	i324m_15
		cmp	dword ptr [_config_cpu_type], 4
		jne	i324m_illegal

		jmp	i324m_end
i324m_15:	cmp	ebx, 006h		; DACR0 (EC040)
		jne	i324m_16
		cmp	dword ptr [_config_cpu_type], 4
		jne	i324m_illegal

		jmp	i324m_end
i324m_16:	cmp	ebx, 007h		; DACR1 (EC040)
		jne	i324m_17
		cmp	dword ptr [_config_cpu_type], 4
		jne	i324m_illegal

		jmp	i324m_end
i324m_17:					; ILLEGAL OPCODE when here
						; ILLEGAL OPCODE when here
i324m_illegal:	
		mov	ebx,010h
		mov	eax,38
		jmp	prepare_exception_

i324m_end:	mov	eax, 4
		ret


;
; Instruction MOVES Rn,<ea> / <ea>,Rn                              ()
; Table use: 0 - routine, 1 - reg
;            2 - arproc,  3 - awproc 
;
; It should use SFC / DFC registers to access memory... Currently, it doesn't


		ALIGN32

public i325_B_
i325_B_:	CPU_LOG	325
		test	byte ptr _sr+1, 20h	; Are we in supervisormode??
		jz	privilegie_violation
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		mov	bx, dx
		rol	bx, 6
		and	ebx, 3ch
		test	dx, 800h
		jz	i325_Br
		mov	edx, [_d + ebx]          ; write to memory
		mov	ebx, [edi + 4]
		call	[edi + 12]
		mov	eax, 4
		ret
i325_Br:	mov	edx, [edi + 4]
		call	[edi + 8]
		test	bl, 8                    ; is this An register ?
		jz	i325_B1
		movsx	edx, dl
		mov	[_d + ebx], edx
		mov	eax, 4
		ret
i325_B1:	mov	byte ptr [_d + ebx], dl
		mov	eax, 4
		ret


		ALIGN32

public i325_W_
i325_W_:	CPU_LOG	325
		test	byte ptr _sr+1, 20h      ; Are we in supervisormode??
		jz	privilegie_violation
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		mov	bx, dx
		rol	bx, 6
		and	ebx, 3ch
		test	dx, 800h
		jz	i325_Wr
		mov	edx, [_d + ebx]          ; write to memory
		mov	ebx, [edi + 4]
		call	[edi + 12]
		mov	eax, 4
		ret
i325_Wr:	mov	edx, [edi + 4]
		call	[edi + 8]
		test	bl, 8                    ; is this An register ?
		jz	i325_W1
		movsx	edx, dx
		mov	[_d + ebx], edx
		mov	eax, 4
		ret
i325_W1:	mov	word ptr [_d + ebx], dx
		mov	eax, 4
		ret


		ALIGN32

public i325_L_
i325_L_:	CPU_LOG	325
		test	byte ptr _sr+1, 20h      ; Are we in supervisormode??
		jz	privilegie_violation
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		mov	bx, dx
		rol	bx, 6
		and	ebx, 3ch
		test	dx, 800h
		jz	i325_Lr
		mov	edx, [_d + ebx]          ; write to memory
		mov	ebx, [edi + 4]
		call	[edi + 12]
		mov	eax, 4
		ret
i325_Lr:	mov	edx, [edi + 4]
		call	[edi + 8]
		mov	[_d + ebx], edx
		mov	eax, 4
		ret


;
; Instruction MULS.L <ea>,Dn / <ea>,Dh:Dl                          (NZV,C=0)
; Table use: 0 - routine, 1 - earegister, 2 - scrproc
;


		ALIGN32

public i326_
i326_:		CPU_LOG	326
		mov	ecx, esi		; Extension word
		fetchwopcode_inc_pos
		add	esi, 2
		mov	ebx, edx		; Get dl nr
		shr	ebx, 10
		and	ebx, 1ch
		mov	eax, dword ptr[_d + ebx]; Get dl value
		push	ebx			; Save dl nr
		mov	ebx, edx
		mov	edx, dword ptr [edi + 4]
		call	dword ptr [edi + 8]	; Get value at <ea>
		test	bh, 8
		jz	i326um
		mov	ebp, 1			; Flag signed operations
		imul	edx			; Signed mul
		jmp	short i326cm
i326um:		xor	ebp, ebp		; Flag unsigned operations
		mul	edx			; Unsigned mul
i326cm:		pop	ecx
		mov	dword ptr [_d + ecx], eax; Low result into dl
		and	byte ptr [_sr], 0f0h
		mov	cl, 0
		test	bh, 4
		jz	i326_3
		and	ebx, 7			; 64 bit result
		mov	dword ptr [_d + 4*ebx], edx
		test	eax, eax
		jnz	i326_1
		test	edx, edx
		jnz	i326_1
		or	cl, 4			; Zero 64 bit result 
i326_1:		test	edx, edx
		jns	i326_2
		or	cl, 8			; Negative 64 bit result
i326_2:		or	byte ptr [_sr], cl
		mov	eax, 4
		ret

i326_3:		test	edx, edx		; 32 bit result
		jz	i326_4			; zero is always OK
		test	ebp, ebp
		jz	i326_7			; Always overflow if unsig
		cmp	edx, 0ffffffffh
		je	i326_4
i326_7:		or	cl, 2			; 32 bit overflowed result
i326_4:		test	eax, eax
		jnz	i326_5
		or	cl, 4			; 32 bit zero result
i326_5:		test	eax, eax
		jns	i326_6			; 32 bit negative result
		or	cl, 8
i326_6:		or	byte ptr [_sr], cl
		mov	eax, 4
		ret


;
; Instruction PACK -(Ax),-(Ay),# / Dx,Dy,#                         ()
; Table use: 0 - routine, 1 - srcear, 2 - scrproc(W)
;            3 - dstear, 4 - dstproc(B)
;


		ALIGN32

public i327_
i327_:		CPU_LOG	327
		mov	edx, [edi + 4]
		call	[edi + 8]
		mov	eax, edx
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		add	edx, eax
		shl	dh, 4
		and	dl, 0fh
		or	dl, dh
		mov	ebx, [edi + 12]
		call	[edi + 16]
		mov	eax, 4
		ret


;
; Instruction PFLUSH A / <fc>,# / <fc>,#,<ea>                      ()
; Instruction PLOADR <fc>,<ea> / PLOADW <fc>,<ea>                  ()
; Instruction PMOVE reg,<ea> / <ea>,reg / PMOVEFD <ea>,reg         ()
; Instruction PTESTR/W <fc>,<ea>,# / <fc>,<ea>,#,An                ()
; Table use: 0 - routine, 1 - srcreg, 2 - srcproc
;
; PFLUSH is not and probably will NEVER be really supported (ATC command)
; PLOAD  is not and probably will NEVER be really supported (ATC command)
; PMOVE  is not yet supported (MMU command)
; PTEST  is not and probably will NEVER be really supported (ATC command)
; Note:	Supervisor


		ALIGN32

public MMUIllegal030_
MMUIllegal030_:	CPU_LOG	0
		test	byte ptr [_sr + 1], 020h
		jz	privilegie_violation	; Priv.violation in user mode 
		mov	ebx, 02ch		; F-line in super mode 
		mov	eax, 4
		jmp	prepare_exception_

	
		ALIGN32

public MMU030_
MMU030_:	CPU_LOG	328
		test	byte ptr [_sr + 1], 020h
		jz	privilegie_violation
	
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2		; skip param word
		mov	eax, 4
		ret


		ALIGN32

public PFLUSHEC040_
PFLUSHEC040_:	CPU_LOG 0
		mov	eax, 4			; NOP on EC040
		ret


		ALIGN32

public PTESTEC040_
PTESTEC040_:	CPU_LOG 0
		mov	eax, 4			; NOP on EC040
		ret


;
; Instruction RTD #                                                ()
; Table use: 0 - routine
;


		ALIGN32

public i329_
i329_:		CPU_LOG	329
		mov	ecx, esi
		fetchwopcode_inc_pos
		movsx	eax, dx
		mov	ecx, dword ptr [_a + 28]
		fetchl
#ifdef PREFETCH
		call	prefetch_fill_
#endif
		mov	esi, edx
		PC_TO_PTR esi
		add	ecx, 4
		add	ecx, eax
		mov	dword ptr [_a + 28], ecx
		mov	eax, 4
		ret


;
; Instruction TRAPcc / .W # / .L #                                 ()
; Table use: 0 - routine, 1 - cond handle proc, 2 - mode
;


		ALIGN32

public i330_
i330_:		CPU_LOG	330
		call	[edi + 4]		; edx = 1 succeeded
		test	dl, dl
		jz	i330_n
		cmp	dword ptr [edi + 8], 0
		je	i330_0
		mov	ecx, esi
		cmp	dword ptr [edi + 8], 1
		je	i330_w
		fetchlopcode_inc_pos
		add	esi, 4
		jmp	short i330_0
i330_w:		fetchwopcode_inc_pos
		add	esi, 2
i330_0:		mov	ebx, 1ch		; make exc 7
		mov	eax, 100
		jmp	prepare_exception_

i330_n:		cmp	dword ptr [edi + 8], 0
		je	i330_e
		add	esi, 2
		cmp	dword ptr [edi + 8], 1
		je	i330_e
		add	esi, 2
i330_e:		mov	eax, 4
		ret


;
; Instruction UNPK -(Ax),-(Ay),# / Dx,Dy,#                         ()
; Table use: 0 - routine, 1 - srcear, 2 - scrproc(B)
;            3 - dstear, 4 - dstproc(W)
;


		ALIGN32

public i331_
i331_:		CPU_LOG	331
		mov	edx, [edi + 4]
		call	[edi + 8]
		mov	ah, dl
		shl	ah, 4
		mov	al, dl
		and	al, 0fh
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		add	dx, ax
		mov	ebx, [edi + 12]
		call	[edi + 16]
		mov	eax, 4
		ret


;
; Instruction CALLM #x, <ea>                         
; Table use: 0 - routine, 1 - src reg, 2 - ea calculate rutine
;


;------------------------------------------------------------
; Validates the format of a module descriptor or stack frame
; Input:  edx - opt-type byte
; Output: ebp - opt,  edx - type
;------------------------------------------------------------


		ALIGN32

callm_validate_:mov	edx, ebp		; Check opt 
		shr	ebp, 5
		and	ebp, 7
		test	ebp, ebp
		jz	callm_optOK
		cmp	ebp, 4
		jne	callm_ferr
callm_optOK:	and	edx, 01fh		; Check type
		test	edx, edx
		jz	callm_typOK
		cmp	edx, 1
		jne	callm_ferr
callm_typOK:	ret
		

	
; CALLM format error below
	
callm_ferr:	ret		; Do nothing about it right now

	
		ALIGN32

public i332_
i332_:		CPU_LOG	332
		mov	ecx, esi	; Load argument count
		fetchwopcode_inc_pos
		add	esi, 2
		mov	ebx, edx
		mov	edx, [edi + 4]	; Calulate module descriptor pointer 
		call	dword ptr [edi + 8]
		mov	eax, ecx
		fetchb			; Fetch opt type field 
		call	callm_validate_
		
		mov	eax, 4
		ret


;
; Instruction RTM Rn
; Table use: 0 - routine, 1 - reg
;

; Just advance PC, no action now


		ALIGN32

public i333_
i333_:		CPU_LOG	333
		mov	eax, dword ptr [edi + 4]  ; Module data area ptr 
		mov	ebx, dword ptr [_a + 28]  ; Get stack ptr 
		mov	ecx, ebx		  ; Get opt and type
		fetchb
		call	callm_validate_
		mov	ecx, ebx		  ; Get condition codes 
		add	ecx, 3
		fetchb
		mov	byte ptr [_sr], dl
		mov	ecx, ebx		  ; Update data area ptr 
		add	ecx, 12
		fetchl
		mov	dword ptr [eax], edx
		mov	ecx, ebx		  ; Get saved PC
		add	ecx, 12
		fetchl
#ifdef PREFETCH
		call	prefetch_fill_
#endif
		mov	esi, edx
		PC_TO_PTR esi
		mov	ecx, ebx		  ; Get saved stack ptr 
		add	ecx, 20
		fetchl
		mov	dword ptr [_a + 28], edx  ; Update stack ptr 
		mov	eax, 4
		ret


;
; Instruction MOVE16 (040 060)
;
; Type 1: move16 (Ax)+, (Ay)+   (post-increment format)
;
; Type 2: move16 $L, (An)   11  (absolute long address or destination)
;         move16 $L, (An)+  01
;	  move16 (An), $L   10
;	  move16 (An)+, $L  00    
;
; Table-usage:
;   0 - Routine-pointer  1 - Pointer to address register
;


		ALIGN32

public i334_1_
i334_1_:	CPU_LOG 0
		mov	eax, dword ptr [edi + 4]
		mov	ebx, dword ptr [eax]		; ebx - Src EA
		add	dword ptr [eax], 16
		mov	ecx, esi
		fetchwopcode_inc_pos
		add	esi, 2
		shr	edx, 11
		and	edx, 03ch
		mov	eax, dword ptr [_a + edx]	; eax - Dst EA
		and	eax, 0fffffff0h
		and	ebx, 0fffffff0h
		mov	ebp, 12
i334_1_a:	lea	ecx, [ebx + ebp]
		fetchl
		lea	ecx, [eax + ebp]
		writel
		sub	ebp, 4
		jns	i334_1_a
		mov	eax, 4
		ret


		ALIGN32

public i334_2_00_
i334_2_00_:	CPU_LOG 0
		mov	eax, dword ptr [edi + 4]
		mov	ebx, dword ptr [eax]		; ebx - Src EA
		add	dword ptr [eax], 16
		mov	ecx, esi
		fetchlopcode_inc_pos
		add	esi, 4
		mov	eax, edx			; eax - Dst EA
		and	eax, 0fffffff0h
		and	ebx, 0fffffff0h
		mov	ebp, 12
i334_2_00_a:	lea	ecx, [ebx + ebp]
		fetchl
		lea	ecx, [eax + ebp]
		writel
		sub	ebp, 4
		jns	i334_2_00_a
		mov	eax, 4
		ret


		ALIGN32

public i334_2_01_
i334_2_01_:	CPU_LOG 0
		mov	ebx, dword ptr [edi + 4]
		mov	eax, dword ptr [ebx]		; eax - Dst EA
		add	dword ptr [ebx], 16
		mov	ecx, esi
		fetchlopcode_inc_pos
		add	esi, 4
		mov	ebx, edx			; ebx - Src EA
		mov	ebx, dword ptr [_a + ebx]	; Dst EA
		and	eax, 0fffffff0h
		and	ebx, 0fffffff0h
		mov	ebp, 12
i334_2_01_a:	lea	ecx, [ebx + ebp]
		fetchl
		lea	ecx, [eax + ebp]
		writel
		sub	ebp, 4
		jns	i334_2_01_a
		mov	eax, 4
		ret



		ALIGN32

public i334_2_10_
i334_2_10_:	CPU_LOG 0
		mov	eax, dword ptr [edi + 4]
		mov	ebx, dword ptr [eax]		; ebx - Src EA
		mov	ecx, esi
		fetchlopcode_inc_pos
		add	esi, 4
		mov	eax, edx			; eax - Dst EA
		and	eax, 0fffffff0h
		and	ebx, 0fffffff0h
		mov	ebp, 12
i334_2_10_a:	lea	ecx, [ebx + ebp]
		fetchl
		lea	ecx, [eax + ebp]
		writel
		sub	ebp, 4
		jns	i334_2_10_a
		mov	eax, 4
		ret


		ALIGN32

public i334_2_11_
i334_2_11_:	CPU_LOG 0
		mov	ebx, dword ptr [edi + 4]
		mov	eax, dword ptr [ebx]		; eax - Dst EA
		mov	ecx, esi
		fetchlopcode_inc_pos
		add	esi, 4
		mov	ebx, edx			; ebx - Src EA
		mov	ebx, dword ptr [_a + ebx]	; Dst EA
		and	eax, 0fffffff0h
		and	ebx, 0fffffff0h
		mov	ebp, 12
i334_2_11_a:	lea	ecx, [ebx + ebp]
		fetchl
		lea	ecx, [eax + ebp]
		writel
		sub	ebp, 4
		jns	i334_2_11_a
		mov	eax, 4
		ret


;
; Instruction CINV and CPUSH (040 040LC)
;
; NOP implementation
;
; Table-usage:
;   0 - Routine-pointer  1 - reg ptr,  2 - cache spec,  3 - scope spec
;


		ALIGN32

public CINV040_
CINV040_:	CPU_LOG 0
		mov	eax, 4			; NOP 
		ret


		ALIGN32

public CPUSH040_
CPUSH040_:	CPU_LOG 0
		mov	eax, 4			; NOP
		ret


_TEXT ends

_DATA segment dword public 'DATA'

include data\memory.inc
include data\68000.inc
include data\68030.inc
include data\68kcopr.inc


eac063istab:	dd offset eac063is0, offset eac063is1, offset eac063is3
		dd offset eac063is4, offset eac063is5, offset eac063is6
		dd offset eac063is7

extern _debugging:dword

_DATA ends

_BSS segment dword public 'DATA'
_BSS ends
end
